.\" @(#)snns2c.1 93/07/12
.TH RUNTIME 3  "19 July 1993"
.SH NAME
Init_Net, Load_Pattern_File, Propagate_Pattern_Array, Propagate_Pattern, Write_Pattern_Arrays, Write_Pattern, runtime \- functions to handle a neural network. These functions are located in the file runtime.c that is created by snns2c.
.SH SYNOPSIS
.nf
.ft B
#include <stdio.h>
#include "runtime.h"
.ft
.fi
.LP
.nf
.ft B
int Init_Net ();
.ft
.fi
.LP
.nf
.ft B
int Load_Pattern_File (in_file, in_patterns, out_patterns);
FILE *in_file;
Patterns *in_patterns, *out_patterns;
.ft
.fi
.LP
.nf
.ft B
int Propagate_Pattern_Array (in_patterns, out_patterns);
Patterns *in_patterns, *out_patterns;
.ft
.fi
.LP
.nf
.ft B
int Propagate_Pattern (input_pattern, output_pattern);
float *input_pattern, *output_pattern;
.ft
.fi
.LP
.nf
.ft B
int Write_Pattern_Arrays (out_file, start, in_patterns, out_patterns);
FILE *out_file; 
int start;
Patterns in_patterns, out_patterns;
.ft
.fi
.LP
.nf 
.ft B
int Write_Pattern (out_file, input_pattern, output_pattern);
FILE *out_file;
float *input_pattern, *output_pattern;
.ft
.fi
.LP
.nf
.ft B
const char *err_chk (err);
int err;
.SH DESCRIPTION
.LP
These functions are parts of the file runtime.c that is created by snns2c.
You can use them to propagate patterns through a neural network.
.SH CONSTANTS
The following constants are defined in runtime.h:
.TP 23
.B int NO_OF_UNITS
total number of units in the network
.TP
.B int NO_OF_LINKS
total number of links in the network
.TP
.B int NO_OF_UNIT_TYPES
number of different unit prototypes
.TP
.B int NO_OF_SITE_TYPES
number of different site types
.TP
.B int MIN_UNIT_NO
number of the first unit
.TP
.B int MAX_UNIT_NO
number of the last unit
.TP
.B int NO_OF_IN_UNITS
number of input units. This value is necessary for the use of the
function Propagate_Pattern
.TP
.B int NO_OF_OUT_UNITS
number of output units. This value is necessary for the use of the
function Propagate_Pattern
.TP
.B char *NET_FILE_VERSION
version of the SNNS network file that was used to generate runtime.c
and runtime.h
.TP
.B char *NET_NAME
name of the network as given in the network file.
.TP
.B char *PAT_IN_FILE
proposed name of the pattern file
.TP
.B char *PAT_RES_FILE
proposed name of the result file
.SH USAGE
.LP
Each function returns an integer value as an error code. A value of 0 
means that no error has occured. The integer can be interpreted by the 
function err_chk(), which returns a pointer to a string explaining the error.
.LP
Before using any of the functions the net has to be initialized with
with Init_Net().
.LP
Load_Pattern_File() loads patterns from the file in_file into the
array in_patterns and allocates memory for the corresponding array
out_patterns. An internal variable which determines the number of
patterns is set.
.LP
Propagate_Pattern_Array() propagates in_patterns and stores the
network output patterns in the array out_patterns.
.LP
Propagate_Pattern() propagates the single pattern input_pattern
through the network. The network output is stored into
output_pattern. input_pattern and output_pattern are pointers to 
arrays of float which must be provided by the user. The size of the arrays must
contain at least as many entries as given by NO_OF_IN_UNITS and
NO_OF_OUT_UNITS.
.LP
Write_Pattern_Arrays() writes the in_patterns and out_patterns arrays
to the file given by out_file. The format is similar to the SNNS
result file format. If in_patterns is NULL no input patterns are
written.  If out_patterns is NULL no output patterns are
written. start defines the number of the first written
pattern.
.LP
Write_Pattern() writes a single pattern pair to the result file
out_file.  The pattern is defined like described for the function
Propagate_Pattern(). If input_pattern or output_pattern is NULL, the
corresponding part of the pattern is not written to the file.
.SH "SEE ALSO"
.BR snns2c.
.SH NOTES
These functions represent the version 2.1 of SNNS. All features of
SNNS implemented in later versions are not supported yet.
.LP
runtime.c and runtime.h are only generic names for the files which are
created by snns2c. See snns2c for more details.
.SH BUGS
Propagate_Pattern_Array and Write_Pattern_Array refer to the last
call to Load_Pattern_File to determine the total number of patterns.
.LP
Please report bugs and bug fixes which are of common interest to
snns@informatik.uni-stuttgart.de
.LP
