/**************************************************************************
 FILE           : snns2c.c
 SNNS VERSION   : 2.1

 PURPOSE        : SNNS Network to C Compiler
 NOTES          :

 AUTHOR         : Kais Brahim
 VERSION        : 1.0
 CHANGED BY     : Dietmar Posselt (05.07.93)

 DATE           : 15.11.1992

             Copyright (c) 1990-1993  SNNS Group, IPVR, Univ. Stuttgart, FRG

*****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "snns2c_type.h"    /* type definitions    */
#include "snns2c.h"         /* strings definitions */
#include "snns2c_funcs.tbl" /* functions table     */
#include "snns2c_prot.h"    /* function prototypes */


FILE    *io_stream, *run_c, *run_h;




/*******************************************************************

		      network   parameters

********************************************************************/




int  no_of_units,               /*  no. of units in the network  */
     min_unit_no      = 0,	/*  the first (lowest) used unit number in the network	*/
     max_unit_no      = 0,	/*  the last (highest) used unit number in the network	*/
     no_in_units      = 0,	/*  no. of input units	 */
     no_out_units     = 0,      /*  no. of output units  */
     no_of_links         ,      /*  no. of links         */
     link_entries     = 0,      /*  no. of link entries  */
     no_of_site_types,
     no_of_unit_types;

char NET_FILE_VERSION[20],
     *Netname;

bool Version_3D;






/**************************************************************************

			   network defaults

***************************************************************************/



char  *UPDATE_FUNC,
      LearnFunc[50]     = "Learn_backprop",
      *Def_actfunc,
      *Def_outfunc,
      Def_st            = 'h',
      *func1,
      *func2;

float Def_act           =  0,
      Def_bias          =  0 ;








char *c_head[] ={
"/**************************************************************************",
"   NET NAME    : ",
"   GENERATED AT:",
"   SNNS VERSION: ",
"   PURPOSE     : ",
"   NOTES       : this file is generated by snns2c",
"   UPDATE FUNC : ",
"   LEARN FUNC  : ",
"   AUTHOR      : Kais Brahim\n\
   CHANGED BY  : Dietmar Posselt\n\
   VERSION     : 1.0\n\n\
             Copyright (c) 1990-1993  SNNS Group, IPVR, Univ. Stuttgart, FRG\n\n\
*****************************************************************************/ \n \n"
};




char *cnet_var[] =
{
"#define     NO_OF_UNITS        ",
"#define     NO_OF_LINKS        ",
"#define     NO_OF_UNIT_TYPES   ",
"#define     NO_OF_SITE_TYPES   ",
"#define     MIN_UNIT_NO        ",
"#define     MAX_UNIT_NO        ",
"#define     NO_OF_IN_UNITS      ",
"#define     NO_OF_OUT_UNITS     ",
"#define     NO_OF_LINK_ENTRIES  ",
"#define     DEF_ACT_FUNC       ",
"#define     DEF_OUT_FUNC       ",
"#define     UPDATE_FUNC       ",

};





char    *title[] = {
    "SNNS network definition file",
    "generated at",
    "network name",
    "no. of units",
    "no. of connections",
    "no. of unit types",
    "no. of site types",
    "learning function",
    "site definition section",
    "type definition section",
    "unit definition section",
    "connection definition section",
    "subnet definition section",
    "unit default section",
    "source files",
    "layer definition section",
    "update function",
    "3D translation section",
    "Init_Units()"
    };

int  NoOfTitles = (sizeof title) / (sizeof title[0]);


char    *patHeader[] = {
    "SNNS pattern definition file %s\n",
    "generated at",
    "No. of patterns     : %d\n",
    "No. of input units  : %d\n",
    "No. of output units : %d\n"
    };


char     *headers[] = {
    " site name | site function",
    " name | act func | out func | sites",
    " act | bias | st | subnet | layer | act func | out func",
    " no. | typeName | unitName | act | bias | st | position | act func | out func | sites",
    " target | site | source:weight",
    " subnet | unitNo.",
    " layer | unitNo.",
    " delta x | delta y | z"
    };



/*************************************************************************

                       my_strdup

*************************************************************************/

char *my_strdup (string)
char *string;
{
int i;
char *ret;
ret = (char *) malloc ((1+strlen(string)) * sizeof(char));
strcpy (ret, string);
return ret;
}






/*************************************************************************

			error   function

*************************************************************************/



void err_chk(err)
int err;
{
  switch(err)
  {
   case ERR_NO:  break;

   case ERR_MEMORY:
	printf("\nInsufficient Memory!\n");
	exit(1);

   case ERR_OUTFUNC:
	printf("\nInvalid Output Function!\n");
	exit(1);

   case ERR_ACTFUNC:
	printf("\nInvalid Input Function!\n");
	exit(1);

   case ERR_UPDFUNC:
	printf("\nInvalid Update Function!\n");
	exit(1);

   case ERR_SITEFUNC:
	printf("\nInvalid Site Function!\n");
	exit(1);

   case ERR_IO:
	printf("\nPhysical I/O Error!\n");
	exit(1);

   case ERR_EOF:
	printf("\nUnexpected EOF!\n");
	exit(1);

   case ERR_FILE_FORMAT:
	printf("\nIncompatible File Format!\n");
	exit(1);

   case ERR_FILE_OPEN:
	printf("\nCannot Open File!\n");
	exit(1);

   case ERR_FILE_C_OPEN:
	printf("\nCannot Open File: OUTPUT.C !\n");
	exit(1);

   case ERR_FILE_H_OPEN:
	printf("\nCannot Open File: OUTPUT.H !\n");
	exit(1);

   case ERR_SYNTAX:
	printf("\nSyntax Error!\n");
	exit(1);

   case ERR_TTYPE:
	printf("\nI/O Unit Type Invalid!\n");
	exit(1);

   case ERR_NO_LINKS:
	printf("\nNo. of Links doesn't match!\n");
	exit(1);

   case ERR_NO_SITE_TYPES:
	printf("\nNo. Of Site Types doesn't match!\n");
	exit(1);

   case ERR_NO_UNIT_TYPES:
	printf("\nNo. Of Unit Types  doesn't match!\n");
	exit(1);

   case ERR_NO_UNITS:
	printf("No. Of Units doesn't match!\n");
	exit(1);
 
   case ERR_OLD_VERSION:
	printf("\nOld File Format Version\n");
	exit(1);
   }

}




/***********************************************************************


	          write headers and function definitions 


*************************************************************************/







int write_inchead(date, name)
char *date,
     *name;

{
   if((fprintf(run_h,"%s\n%s%s \n%s %s \n\n%s%s\n%s%s\n%s\n\n%s%s\n%s%s\n%s\n",
	c_head[0], c_head[1], name, c_head[2], date,
	c_head[3], NET_FILE_VERSION,
	c_head[4], "constant declarators and default values", c_head[5], 
        c_head[6], UPDATE_FUNC, c_head[7], LearnFunc, c_head[8]))==EOF)
   return ( 1 );

   return(0);
}



int write_chead(date,name)
char *date,
     *name;
{
   if((fprintf(run_c,"%s\n%s%s \n%s %s\n",c_head[0], c_head[1],
      name, c_head[2], date))==EOF)
   return( 1 );

   if((fprintf(run_c,"%s%s\n%s%s%s\n\n%s%s\n%s%s\n\n%s",
        c_head[3], NET_FILE_VERSION,
	c_head[4], "contain the network initializations and pattern\n\
                 propagating functions\n", c_head[5], c_head[6],
                  UPDATE_FUNC, c_head[7], LearnFunc, c_head[8]))==EOF)
   return ( 1 );

   return (0);
}



int write_cinclude()
{
  char *tmp1, *tmp2;

  if((fprintf(run_c,"%s\n\n\n",c_include))==EOF) return(1);

  tmp1 = my_strdup (Def_actfunc);
  tmp2 = my_strdup (Def_outfunc);

  if(!set_func(&tmp1)) err_chk(ERR_ACTFUNC);
  if(!set_func(&tmp2)) err_chk(ERR_OUTFUNC);

  if(fprintf(run_c,"%s(TransFuncPtr)%s\n%s(TransFuncPtr)%s\n\n\n\n",
      cnet_var[9], tmp1, cnet_var[10], tmp2)==EOF)
  return(1);

  return(0);
}

int write_macros_and_types()
{  
  FILE *file;
  int ch;
 
  if((file = fopen(funcs_file0, "r")) == NULL)
   {
   printf ("\nCannot Open File: %s!\n", funcs_file0);
   exit(1);
   }

  while ((ch=fgetc(file)) != EOF)
   {
   if (fputc(ch, run_c)==EOF)
    return(1);
   }
  return(0);
}

int write_prototypes()
{
  if((fprintf(run_c,"%s\n\n\n",prototypes))==EOF) return(1);

  return(0);
}


int write_def()
{

   if((fprintf(run_h,"#define     NET_FILE_VERSION    \"%s\"\n\
#define     NET_NAME           \"%s\"\n\
#define     PAT_IN_FILE        \"%s.pat\"\n\
#define     PAT_RES_FILE       \"%s.res\"\n\n\n%s\n\n\n",
	NET_FILE_VERSION, Netname, Netname, Netname, export_prototypes))==EOF) return(1);

   return(0);

}

int write_initnet()
{
  if((fprintf(run_c,"%s\n", initnet_func))==EOF) return(1);
  if(no_of_site_types)
	  { if((fprintf(run_c,"%s\n", sitetype_call))==EOF) return(1);}
  if(no_of_unit_types)
	  { if((fprintf(run_c,"%s\n", unittype_call))==EOF) return(1);}

  return(0);
}


int write_cvart()
{
  if((fprintf(run_h,"%s%d\n%s%d\n%s%d\n%s%d\n%s%d\n%s%d\n%s%d\n%s%d\n%s%d\n\n\n\n",
     cnet_var[0], no_of_units, cnet_var[1], no_of_links,
     cnet_var[2], no_of_unit_types, cnet_var[3],  no_of_site_types,
     cnet_var[4], min_unit_no, cnet_var[5], max_unit_no,
     cnet_var[6], no_in_units, cnet_var[7], no_out_units,
     cnet_var[8], link_entries))==EOF)
  return(1);

  if((fprintf(run_h,"%s(FunctionPtr)%s\n\n\n\n",
      cnet_var[11], UPDATE_FUNC))==EOF)
  return(1);

  return(0);
}


int write_cvar()
{
  if((fprintf(run_c,"%s", c_var))==EOF)  return(1);

  return(0);
}





/************************************************************************


			   scanning functions


*************************************************************************/







bool    SkipComments()
{
  register int   c;
  register  FILE  *stream;


  stream = io_stream;

  while (TRUE)
    {
    do      c = getc( stream );
    while ( isspace( c ) );

    if (c != '#')  break;

    do
      c = getc( stream );
    while (c != EOF && c != '\n');

  }

  if (c == EOF)  return( FALSE );

  ungetc( c, stream );
  return( TRUE );
}







bool     MatchComma()
{
  register int  c;
  register  FILE  *stream;


  stream = io_stream;

  do    c = getc( stream );
  while ( isspace( c ) );

  if (c == EOF)  err_chk(EOF);
  if (c != ',')
    {
    ungetc( c, stream );
    return( FALSE );
    }

  do
    {
    c = getc( stream );
    if (c == '\n') {
		    if( no_of_site_types == 0)
		    fprintf(run_c,"\n");
		   }
    }
  while ( isspace( c ) );

  ungetc( c, stream );
  return( TRUE );
}






bool GetNewLine(io_st)
FILE * io_st;
{
  register int  c;

  do
    {
    c = getc( io_st );
    if (c == '\n')  return( TRUE );
    }
  while ( isspace( c ) );

  if (c == EOF)  err_chk(ERR_EOF);

  ungetc( c, io_st);
  return( FALSE );
}










bool     SkipSepar()
{
  register int  c;

  do
    {
    c = getc( io_stream );
    if (c == '\n')  return( FALSE );
    }
  while ( isspace( c ) );

  if (c == EOF)  err_chk(ERR_EOF);
  if (c == '|')
    return( TRUE );

  ungetc( c, io_stream );
  return( FALSE );
}








char     *getSection( line, title_no , Title, NoOfT, io_stream)
FILE    *io_stream;
char    *line, *Title[];
int     *title_no, *NoOfT;
{
  register int   i,j = 0,k = 0, l ;
  register char  *str1, *str2;

  if ( !SkipComments() )
    {  /* EOF  */
    *title_no = -2;
    return( NULL );
    }

  if (fgets( line, LIN_MAX, io_stream ) == NULL)
    {  /*  I/O error  */
    *title_no = -3;
    return( NULL );
    }


  for (i = 0; i < *NoOfT; i++)
    {
    str1 = line + j; while (*str1 == ' '){str1++;}
    str2 = Title[i] +j;  while (*str2 == ' ') {str2++;}


    while (*str2 != EOS)
      {  /* compare strings, don't examine white space  */
      while (isspace(*str1)) {str1++;j++;}
      while(isspace(*str2)) str2++;
      if (*str1++ != *str2++)
      { k = j; break;}
      else j++;

      if (*str2 == EOS)
	  {
	  if (k > 0)
	  {
	  str1 = line;
	  str2 = Title[i];
	  for(l=0; l<= k; l++)
	  if (*str1++ != *str2++) {j = k; break;}
	  str1 = line + j + 1;
	  }

	  /*  strings are equal   */
	  *title_no = i;
	  return( str1 );
	  }
      }
    }


  *title_no = -1;  /*  title not found  */
  return( NULL );
}





bool    SkipLine( N )
int     N;
{
  register  FILE  *stream;
  register int i, c;



  do    c = getc( io_stream ); while ( isspace( c ) );

  if (c == EOF) err_chk(ERR_EOF);
  ungetc( c, io_stream );

  stream = io_stream;

  c = getc( stream );
   if (c != '-')
     {
     ungetc( c, stream );
     return( FALSE );
     }

  for (i = 0; i < N; i++)
    {
    c = getc( stream );
    if (c != '-') return( FALSE );
    do c = getc( stream );  while ( c == '-' );
    if (c != '|')  err_chk(ERR_SYNTAX);  
    }

  c = getc( stream );
  if (c != '-')  err_chk(ERR_SYNTAX);

  do    c = getc( stream );  while ( c == '-' );

  if (c == EOF)  err_chk(ERR_EOF);
  ungetc( c, stream );

  if ( !GetNewLine(io_stream) )  return( FALSE );

  return( TRUE );
}



bool set_func(func_name)
char **func_name;

{
  int i;
  struct FuncTable *f_ptr;

  for(i = 0, f_ptr = func_table; i< NO_OF_FUNCS; i++, f_ptr++)
   {
    if(!strcmp(f_ptr->func_name, *func_name))
      {
      *func_name = (char *)my_strdup(f_ptr->function_name);
      return(TRUE);
      }
    }

  return(FALSE);
}











/******************************************************************************


			       parsing functions


*****************************************************************************/







void ReadDefaults( stream )
FILE    *stream;
{
  int       ret, subnet_no, layer_no;

  io_stream = stream;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[2] ) != 0)
      err_chk(ERR_SYNTAX);
  if ( !SkipLine( 6 ) )
      err_chk(ERR_SYNTAX);


/* set defaults */

  ret = fscanf( stream, "%f | %f | %c | %d | %d | %s | %s",
		&Def_act, &Def_bias, &Def_st, &subnet_no, &layer_no, Def_actfunc, Def_outfunc );
  if (ret != 7) err_chk(ERR_SYNTAX);

  if(!set_func(&Def_actfunc)) err_chk(ERR_ACTFUNC);
  if(!set_func(&Def_outfunc)) err_chk(ERR_OUTFUNC);


  if((Def_st != 'i')&&(Def_st != 'o')&&(Def_st != 'h')&&
       (Def_st != 'd')&&(Def_st != 's'))
	err_chk(ERR_TTYPE);

  if ( !SkipLine( 6 ) ) err_chk(ERR_SYNTAX);

}






void ReadUnitDefinitions( stream)
FILE    *stream;
{
  float     fp;
  int   no,  st,  ret, ret1, ret2, dummy, NO_OF_UNITS = 0;
  bool  F_type;
  char Input[LIN_MAX];

  io_stream = stream;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[3] ) != 0) err_chk(ERR_SYNTAX);
  if ( !SkipLine( 9 ) ) err_chk(ERR_SYNTAX);

  if(fprintf(run_c,"%s", unit_func)==EOF) err_chk(ERR_IO);



  while (TRUE)
    {
    if ( SkipLine( 9 ) )
      {
       fprintf(run_c,"\n}\n\n\n");
       if(NO_OF_UNITS != no_of_units) err_chk(ERR_NO_UNITS);
       SkipComments();
       return;
       }
    if ( !SkipComments() )  return;

    NO_OF_UNITS++;
    F_type = FALSE;


/*  unit no */

    ret = fscanf( stream, "%d", &no );
    if (ret != 1)  err_chk(ERR_SYNTAX);

    if(min_unit_no == 0) { min_unit_no =no; max_unit_no =no;}
    if (no < min_unit_no) min_unit_no = no;
    if (no > max_unit_no) max_unit_no = no;



/*  Ftype Name */
    *Input = EOS;
    ret = fscanf( stream, " | %[^| ]", Input );

    if (ret != 0 && ret != 1)  err_chk(ERR_SYNTAX);
    if (ret == 1)
     {
	F_type = TRUE;
	if(fprintf(run_c,"\n   Set_Unit(%d,\"%s\"",no,Input)==EOF) err_chk(ERR_IO);
      }
    else if (fprintf(run_c,"\n   Set_Unit(%d, NULL",no)==EOF) err_chk(ERR_IO);


/*  unitName */

    ret = fscanf( stream, " | %[^| ]", Input );
    if (ret == 1 )
       {
       if(fprintf(run_c,", \"")==EOF) err_chk(ERR_IO);
       dummy = 0;
       while((Input[dummy] != EOS)&&(Input[dummy] != '\\')&&(Input[dummy] != '\"'))
	   fprintf(run_c,"%c", Input[dummy++]);
       if(fprintf(run_c,"\"")==EOF) err_chk(ERR_IO);
       }
    else  if(fprintf(run_c,", NULL")==EOF) err_chk(ERR_IO);

/*  unit activation */

    ret = fscanf( stream, " | %f", &fp );
    if (ret == 1)  { if(fprintf(run_c,", %f",fp)==EOF) err_chk(ERR_IO);}
    else  if(fprintf(run_c,", %f", Def_act)==EOF) err_chk(ERR_IO);

/*  unit bias */
    ret = fscanf( stream, " | %f", &fp );
    if (ret == 1)  { if(fprintf(run_c,", %f",fp)==EOF) err_chk(ERR_IO);}
    else  if(fprintf(run_c,", %f", Def_act)==EOF) err_chk(ERR_IO);


/*  unit st */
    ret = fscanf( stream, " | %[^| ]", Input );
    if (ret == 1)
      {
      if(fprintf(run_c,", (int) '%s'",Input)==EOF) err_chk(ERR_IO);
      st = Input[0];
      }
    else
      {
      if(fprintf(run_c,", (int) '%s'", Def_st)==EOF) err_chk(ERR_IO);
      st = Def_st;
      }

    switch(st)
     {
      case'i': no_in_units++;
	       break;
      case'o': no_out_units++;
	       break;
      default: break;
     }

/* Version check */

    if (Version_3D)
/* 3D Version*/
      {
      ret = fscanf( stream, " | %d , %d, %d", &dummy, &dummy, &dummy );
      if (ret != 0 && ret != 3)  err_chk(ERR_OLD_VERSION);
      }
    else
      {
      ret = fscanf( stream, " | %d , %d", &dummy, &dummy );
      if (ret != 0 && ret != 2)  err_chk(ERR_OLD_VERSION);
      }


/*  default type */
    if (F_type)
      {
      ret = fscanf( stream, " | | |" );
      if ( ret != 0 || !GetNewLine(io_stream) )  err_chk(ERR_SYNTAX);
      fprintf(run_c,",NULL,NULL,NULL);");
      continue;
      }


/*  activation function name */
    ret = fscanf( stream, " | %[^| ]", func1 );
    if(ret == 1)
      {
      if(!set_func(&func1)) err_chk(ERR_ACTFUNC);
      fprintf(run_c,", %s", func1);
      }
    else fprintf(run_c,", NULL");

/*  output  function name*/
    ret = fscanf( stream, " | %[^| ]", func2 );
    if(ret == 1)
      {
      if(!set_func(&func2)) err_chk(ERR_OUTFUNC);
      fprintf(run_c,", %s", func2);
      }
    else  fprintf(run_c,", NULL");



    if ( !SkipSepar() )  err_chk(ERR_SYNTAX);
    if ( GetNewLine(io_stream) )  {fprintf(run_c,",NULL);");continue;}



/*  site names  */

    while (TRUE)
      {
      ret = fscanf( stream, " %[^,\n ]", Input );
      if(ret == 1) fprintf(run_c,", \"%s\"",Input);
	else  err_chk(ERR_SYNTAX);
      if ( !MatchComma() )  {fprintf(run_c,",NULL"); break;}
      }
     fprintf(run_c,");");
    }

}





void ReadConnectionDefs( stream )
FILE    *stream;
{
  float  weight;
  int   no, ret, src_unit, NO_OF_LINKS = 0;
  bool  has_sites;
  char Input[LIN_MAX];

  io_stream = stream;
  has_sites = FALSE;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[4] ) != 0) err_chk(ERR_SYNTAX);
  if ( !SkipLine( 2 ) )  err_chk(ERR_SYNTAX);

  if(fprintf(run_c,"%s", conn_func)==EOF) err_chk(ERR_IO);

  while (TRUE)
    {
    if ( SkipLine( 2 ) )
      {
       if(no_of_links != NO_OF_LINKS) err_chk(ERR_NO_LINKS);
       fprintf(run_c,"\n\n}\n\n");
       return;
      }
    if ( !SkipComments() )  return;

    link_entries++;


/*  unit target no */

    ret = fscanf( stream, "%d", &no );
    if(fprintf(run_c,"\n  Set_Links(&link_ptr, %d,", no)==EOF) err_chk(ERR_IO);

    if (ret != 0 && ret != 1)  err_chk(ERR_SYNTAX);

    if (ret == 1)
     {
      has_sites = FALSE;
      if ( !SkipSepar() )  err_chk(ERR_SYNTAX);
     }

/*  site name */
    *Input = EOS;

    ret = fscanf( stream, " %[^| ] ", Input );


    if (ret != 0 && ret != 1) err_chk(ERR_SYNTAX);
    if (has_sites && ret == 0)   err_chk(ERR_SYNTAX);
    if (ret == 1)
       {
	has_sites = TRUE;
	if(fprintf(run_c,"\"%s\",",Input)==EOF) err_chk(ERR_IO);
       }
    if (ret == 0)
     { if(fprintf(run_c," NULL,")==EOF) err_chk(ERR_IO); }

    fprintf(run_c,"\n");

    if ( !SkipSepar() )  err_chk(ERR_SYNTAX);



/* link entry src:weight*/

    while (TRUE)
      {
      ret = fscanf( stream, "%d : %f", &src_unit, &weight );

      if (ret != 2) err_chk(ERR_SYNTAX);
	 else  NO_OF_LINKS++;

      fprintf(run_c,"\"%d:%f\",",src_unit, weight);

      if ( !MatchComma() ) {fprintf(run_c," NULL);\n");break;}

      }

    }
 }












void ReadSiteDefinitions( stream )
FILE    *stream;
{
  int   ret, NO_OF_SITE_TYPES = 0;
  char  str[50];

  io_stream = stream;


  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[0] ) != 0) err_chk(ERR_IO);
  if ( !SkipLine( 1 ) ) err_chk(ERR_SYNTAX);


  if(fprintf(run_c,"%s", site_func)==EOF) err_chk(ERR_IO);

  while (TRUE)
    {
    if (SkipLine( 1 ))
      {
       if(NO_OF_SITE_TYPES != no_of_site_types) err_chk(ERR_NO_SITE_TYPES);
       if(fprintf(run_c,"\n\n}")==EOF) err_chk(ERR_IO);
       return;
      }
    if ( !SkipComments() ) err_chk(ERR_SYNTAX);


    ret = fscanf( stream, "%s | %s\n", str, func1 );
    if (ret != 2) err_chk(ERR_SYNTAX);
    else
      {
       if(!set_func(&func1)) err_chk(ERR_SITEFUNC);
       if(fprintf(run_c,"    Set_Site(&site_ptr, \"%s\", %s);\n", str, func1)==EOF)
       err_chk(ERR_IO);
      }
    NO_OF_SITE_TYPES++;
    }

}








void ReadTypeDefinitions( stream )
FILE    *stream;
{
  int   no_sites,  ret,  i, NO_OF_UNIT_TYPES = 0;
  char str[50]; /*, func1[50], func2[50];  */

  io_stream = stream;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[1] ) != 0)  err_chk(ERR_IO);
  if ( !SkipLine( 3 ) )  err_chk(ERR_SYNTAX);

  if(fprintf(run_c,"\n\n\n%s", utype_func)==EOF) err_chk(ERR_IO);

  while (TRUE)
    {
     if ( SkipLine( 3 ) )
      {
       if(NO_OF_UNIT_TYPES != no_of_site_types) err_chk(ERR_NO_UNIT_TYPES);
       if(fprintf(run_c,"\n}\n\n")==EOF) err_chk(ERR_IO);
       return;
      }
    if ( !SkipComments() )  err_chk(ERR_SYNTAX);

    ret = fscanf( stream, "%s | %s | %s |", str, func1, func2 );

    if (ret != 3) err_chk(ERR_SYNTAX);
    else
     {
     if(!set_func(&func1)) err_chk(ERR_ACTFUNC);
     if(!set_func(&func2)) err_chk(ERR_OUTFUNC);

     if(fprintf(run_c,"    Set_FuncType(F_ptr, \"%s\", %s , %s); \n",
			str, func1, func2)==EOF) err_chk(ERR_SYNTAX);
     }
/* Sites */

    if (!GetNewLine(io_stream) )
     {

    for (no_sites=0; no_sites < no_of_site_types; no_sites++)
      {
       ret = fscanf( stream, "%[^,\n ]", str);
       if(ret == 1)
	     fprintf(run_c,"    Set_SiteEntry(F_ptr,\"%s\");\n", str);
       else  err_chk(ERR_SYNTAX);

       if ( !MatchComma() )  break;
      }
     }

    NO_OF_UNIT_TYPES++;
    fprintf(run_c,"\n\n    F_ptr++;\n\n");
   }

}





void SkipSubnetDefs( stream )
FILE    *stream;
{
  int   unit_no, subnet_no, ret;


  io_stream = stream;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[5] ) != 0) err_chk(ERR_SYNTAX);
  if ( !SkipLine( 1 ) )  err_chk(ERR_SYNTAX);


  while (TRUE)
    {
    if (SkipLine( 1 )) return;
    if ( !SkipComments() )  return;
    ret = fscanf( stream, "%d", &subnet_no );
    if (ret != 1)  err_chk(ERR_SYNTAX);

    if ( !SkipSepar() )  err_chk(ERR_SYNTAX);

    while (TRUE)
      {
      ret = fscanf( stream, "%d", &unit_no );
      if (ret != 1)  err_chk(ERR_SYNTAX);
      if ( !MatchComma() )  break;
      }
    }
}




void SkipLayerDefs( stream )
FILE    *stream;
{
  int   unit_no, layer_no, ret;


  io_stream = stream;

  if ( !SkipComments() )  return;
  if (fscanf( stream, headers[6] ) != 0) err_chk(ERR_SYNTAX);
  if ( !SkipLine( 1 ) ) err_chk(ERR_SYNTAX);

  while (TRUE)
    {
    if (SkipLine( 1 )) return;
    if ( !SkipComments() )  return;
    ret = fscanf( stream, "%d", &layer_no );
    if (ret != 1)  err_chk(ERR_SYNTAX);

    if ( !SkipSepar() )  err_chk(ERR_SYNTAX);

    while (TRUE)
      {
      ret = fscanf( stream, "%d", &unit_no );
      if (ret != 1)  err_chk( ERR_SYNTAX);

      if ( !MatchComma() )  break;
      }
    }
}







/****************************************************************************

			 parsing  network


*****************************************************************************/




void ScanHeader( stream)
FILE    *stream;
{
  char  *cursor, *date, sc, *v_str, Input[LIN_MAX];
  int   err,  i = 0,
	no_of_scan_params,
	title_no;
  time_t  clock;

  io_stream = stream;
  err = 1;


  /*  get file version  */
  if ( (cursor = getSection( Input, &title_no, title, &NoOfTitles, io_stream)) == NULL)  return;

  /*  "SNNS network definition file" not found  */
  if (title_no != 0)  err_chk(ERR_FILE_FORMAT);


  err = sscanf( cursor, "%s", NET_FILE_VERSION );
  /*  version string not found */
  if (err != 1)  err_chk(ERR_FILE_FORMAT);


  v_str = (char *) my_strdup(NET_FILE_VERSION);
  while(*v_str++ != '-');
  if(strcmp(v_str, VERSION_3D))
     {
     Version_3D = TRUE;
     no_of_scan_params = 9;     /* for Version V1.4-3D */
     }
     else
      {
      Version_3D = FALSE;
     no_of_scan_params = 8;     /* for Old Version */
      }

/*  free(v_str); */



  for (i = 0; i < no_of_scan_params; i++)
    {
    if ( (cursor = getSection( Input, &title_no, title, &NoOfTitles, io_stream )) == NULL)  return;

    switch (title_no)
      {
      case  -3:  /*  Physical I/O error  */
	err_chk(ERR_IO);
	return;
      case  -2:  /*  unexpected EOF  */
	err_chk(ERR_EOF);
	return;
      case  -1:  /*  incompatible fileformat */
	err_chk(ERR_FILE_FORMAT);
	return;

      case  1:  /*  "generated at"  */
	 break;

      case  2:  /*  network name"  */
	err = sscanf( cursor, " :%s", Netname );
	if (err == 0) {
	  *Netname = EOS;
	  err = 1;
	}
	break;
      case  3:  /*  "no. of units"  */
	err = sscanf( cursor, " :%d", &no_of_units );
	break;
      case  4:  /*  "no. of connections"  */
	err = sscanf( cursor, " :%d", &no_of_links );
	break;
      case  5:  /*  "no. of unit types"  */
	err = sscanf( cursor, " :%d", &no_of_unit_types );
	break;
      case  6:  /*  "no. of site types"  */
	err = sscanf( cursor, " :%d", &no_of_site_types );
	break;
      case  7:  /*  "learning function" */
	sscanf( cursor, " :%s\n", LearnFunc );
	break;

      case 14:  /*  "source files"  */
	break;
      case  16:  /*  "update function"  */
	/*sscanf( cursor, " :%s\n", UPDATE_FUNC );*/
	break;

      default:
	err_chk(ERR_FILE_FORMAT);  /*  Incompatible file format  */
	return;
     }
   }



   clock = 1;
   (void) time( &clock);


   /* write file headers in runtime.c and runtime.h*/

   if( write_chead(ctime(&clock), Netname) != 0) err= 2;
   if( write_inchead(ctime(&clock), Netname) != 0) err= 2;
   if( write_cinclude() != 0) err = 2;
   if( write_macros_and_types() != 0) err = 2;
   if( write_prototypes() != 0) err = 2;

   /* write global variables in runtime.c */

   if( write_cvar() != 0) err = 2;


   set_func(&UPDATE_FUNC);

   switch (err)
     {
     case 1:    /*  one argument matched  */
       break;
     case 2:
       err_chk(ERR_IO);
       break;
     case EOF:  /*  unexpected EOF  */
       err_chk(ERR_EOF);
       break;
     default:   /*  Syntax error at line  */
       err_chk(ERR_SYNTAX);
       return;
     }


}







void    ScanGenNet( filename)
char    *filename;
{
  FILE *snns_in;
  int  title_no, state;
  char Input[LIN_MAX];

  state = 0;

  snns_in = fopen(filename, "r" );
  if (snns_in == NULL)
    err_chk(ERR_FILE_OPEN);  /*  Can't open file  */



/* scan header */

  ScanHeader( snns_in);

  if(fprintf(run_c,"%s", net_init)==EOF) err_chk(ERR_IO);

  while ( getSection( Input, &title_no, title, &NoOfTitles, io_stream ) != NULL )
    {

    switch (title_no)
      {
      case  -3:  /*  Physical I/O error  */
	err_chk(ERR_IO);
      case  -2:  /*   EOF  */
	if (state < 2)  err_chk(ERR_EOF);  /*  unexpected EOF  */
      case  -1:  /*  not found  */
	err_chk(ERR_SYNTAX);  /*  Syntax error at line  */

      case  8:  /*  "site definition section" */
	ReadSiteDefinitions( snns_in );
	break;

      case  9:  /*  "type definition section"  */
	ReadTypeDefinitions( snns_in );
	state++;
	break;

      case  10:  /*  "unit definition section"  */
	ReadUnitDefinitions( snns_in);
	state++;
	break;

      case  11:  /*  "connection definition section"  */
	ReadConnectionDefs( snns_in );
	if(write_cvart() !=0) err_chk(ERR_IO);
	state++;
	break;


      case  12:  /*  "subnet definition section"  */
	SkipSubnetDefs( snns_in );
	break;


      case  13:  /*  "unit default section"  */
	ReadDefaults( snns_in );
	break;

      case  15:  /*  "layer definition section"  */
	SkipLayerDefs( snns_in );
	break;

      default:
	err_chk(ERR_FILE_FORMAT);  /*  Incompatible file format  */
      }

    }

    if(write_initnet() != 0) err_chk(ERR_IO);
    if(write_def() != 0)  err_chk( ERR_IO);



  fclose(snns_in);

}





bool    getEntry( line)
char    *line;

{
  register int   i,j = 0,k = 0, l ;
  register char  *str1, *str2;


  while (*line == ' ') line++;

  for (i = 0; i < NoOfTitles; i++)
    {
    str1 = line + j; while (*str1 == ' ')str1++;
    str2 = title[i] +j;  while (*str2 == ' ') str2++;


    while (*str2 != EOS)
      {  /* compare strings, don't examine white space  */
      while (*str1 == ' ') {str1++;j++;}
      while (*str2 == ' ') str2++;
      if (*str1++ != *str2++)
      { k = j; break;}
      else j++;

      if (*str2 == EOS)
	  {
	  if (k > 0)
	  {
	  str1 = line;
	  str2 = title[i];
	  for(l=0; l<= k; l++)
	  if (*str1++ != *str2++) {j = k; break;}
	  str1 = line + j + 1;
	  }

	  /*  strings are equal   */
	  return( TRUE );
	  }
      }
    }


  return(FALSE );
}



/****************************************************************************

                    append C functions from snns2c.src to runtime.c

****************************************************************************/


void AppendFuncs(file)
char file[50];
{
  FILE *c_in;
  char Input[LIN_MAX];
  int i;

  c_in = fopen(file, "r");
  if(c_in == NULL) err_chk(ERR_IO);

  do    fgets(Input, 250, c_in);
  while (!getEntry( Input));

  do  fprintf(run_c,"%s",Input);
  while(fgets(Input, 250, c_in));

  fclose(c_in);

}





/***************************************************************************

                       main function

****************************************************************************/






void main(argc, argv)
int argc;
char *argv[];
{
   char  *file,   /* net name */
         *outfile_base,
         *outfile,
	 *str;
   FILE *f_ptr;

   UPDATE_FUNC = my_strdup ("Serial_Order");
   func1 = my_strdup ("");
   func2 = my_strdup ("");
   Def_actfunc = my_strdup ("Act_Logistic");
   Def_outfunc = my_strdup ("Out_Identity");


   printf("\n\tSNNS to C Network Compiler\n\n");

   switch (argc)
     {
     case 2:
       file=my_strdup(argv[1]);
       outfile_base=my_strdup(runtime_c);
       break;
     case 4:
       file=my_strdup(argv[1]);
       outfile_base=my_strdup(argv[3]);
       if (strcmp(argv[2],"-o")==0)
           break;
     default:
       printf ("USAGE: %s <infile.net> [ -o outfile.basename ]\n\n", argv[0]);
       exit (1);
       break;
     }

   if ((file==NULL) || (outfile_base==NULL))
    err_chk(ERR_MEMORY);

   if ((Netname=my_strdup(file))==NULL)
    err_chk(ERR_MEMORY);

   if ((outfile=(char *) malloc(strlen(outfile_base)+3,sizeof(char))) == NULL)
    err_chk(ERR_MEMORY);

   strcpy (outfile, outfile_base);
   strcat (outfile, ".c");

   if((run_c = fopen(outfile, "w")) == NULL)
    err_chk(ERR_FILE_C_OPEN);

   outfile[strlen(outfile)-1] = 'h'; 

   if((run_h = fopen(outfile, "w")) == NULL)
    err_chk(ERR_FILE_H_OPEN);



   ScanGenNet(file);
   AppendFuncs(funcs_file1);

   fclose(run_c);
   fclose(run_h);

   printf("\nNo. of Units: %d\nNo. of Connections: %d\nNo. of Unit Types: %d\n\
No. of Site Types: %d\nLearn Function: %s\nUpdate Function: %s\n", no_of_units, 
no_of_links, no_of_unit_types, no_of_site_types, LearnFunc, UPDATE_FUNC);

   exit(0);

}


